@@ -7,7 +7,7 @@ from jupyterhub.handlers import LogoutHandler
 from jupyterhub.utils import maybe_future
 from oauthenticator.generic import GenericOAuthenticator
 from tornado import gen, web
-from traitlets import Unicode, Set, Bool, Any
+from traitlets import Unicode, Set, Bool, Any, List
 import jwt, os, pwd, time, json
 from urllib import request, parse
 from urllib.error import HTTPError
@@ -62,20 +62,29 @@ class KeyCloakAuthenticator(GenericOAuthenticator):
         help="Users with this role login as jupyterhub administrators"
     )
         
-    get_uid_hook = Any(
-        allow_none=False,
+    pre_spawn_hook = Any(
         help="""
-        Mandatory function to retrieve the user uid from its auth state and pass it to spawner.
+        Function to execute before spawning a session. Usefull to inject extra variables, like the oauth tokens
         Example::
-            def get_uid_hook(spawner, auth_state):
-                spawner.user_uid = auth_state['oauth_user']['cern_uid']
-            c.KeyCloakAuthenticator.get_uid_hook = get_uid_hook
+            def pre_spawn_hook(authenticator, spawner, auth_state):
+                spawner.environment['ACCESS_TOKEN'] = auth_state['exchanged_tokens']['eos-service']
+            c.KeyCloakAuthenticator.pre_spawn_hook = pre_spawn_hook
         """
     ).tag(config=True)
 
+    exchange_tokens = List (
+        Unicode(),
+        default_value=[],
+        config=True,
+        help="List of audiences to exchange our token to"
+    )
+
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
 
+        # Force auth state so that we can store the tokens in the user dict
+        self.enable_auth_state = True
+
         if not self.oidc_issuer:
             raise Exception('No OIDC issuer url provided')
 
@@ -104,7 +113,7 @@ class KeyCloakAuthenticator(GenericOAuthenticator):
     def _decode_token(self, token):
         return jwt.decode(token, verify=False, algorithms='RS256')
 
-    def _get_roles_for_token(self, token):
+    def get_roles_for_token(self, token):
         decoded_token = self._decode_token(token)
         return set(
             decoded_token.\
@@ -113,25 +122,63 @@ class KeyCloakAuthenticator(GenericOAuthenticator):
                get('roles', 'no_roles')
         )
 
+    def _exchange_tokens(self, token):
+
+        tokens = dict()
+
+        for new_token in self.exchange_tokens:
+
+            values = dict(
+                grant_type = 'urn:ietf:params:oauth:grant-type:token-exchange',
+                client_id = self.client_id,
+                client_secret = self.client_secret,
+                subject_token = token,
+                audience = new_token,
+                requested_token_type = 'urn:ietf:params:oauth:token-type:access_token'
+            )
+            data = parse.urlencode(values).encode('ascii')
+
+            req = request.Request(self.token_url, data)
+            with request.urlopen(req) as response:
+                data = json.loads(response.read())
+                tokens[new_token] = data.get('access_token', None)
+
+        return tokens
+
+    def _refresh_token(self, refresh_token):
+        values = dict(
+            grant_type = 'refresh_token',
+            client_id = self.client_id,
+            client_secret = self.client_secret,
+            refresh_token = refresh_token
+        )
+        data = parse.urlencode(values).encode('ascii')
+
+        req = request.Request(self.token_url, data)
+        with request.urlopen(req) as response:
+            data = json.loads(response.read())
+            return (data.get('access_token', None), data.get('refresh_token', None))
+
     async def authenticate(self, handler, data=None):
         user = await super().authenticate(handler, data=data)
-        if user:
-            user_roles = self._get_roles_for_token(user['auth_state']['access_token'])
-            if not self._validate_roles(user_roles):
-                return None
-            user['admin'] = self.admin_role and (self.admin_role in user_roles)
-            self.log.info("Authentication Successful for user: %s, roles: %s, admin: %s" % (user['name'],user_roles,user['admin']))
-            return user
-        else:
+        if not user:
             return None
 
+        user_roles = self.get_roles_for_token(user['auth_state']['access_token'])
+        if not self._validate_roles(user_roles):
+            return None
+
+        user['auth_state']['exchanged_tokens'] = self._exchange_tokens(user['auth_state']['access_token'])
+
+        user['admin'] = self.admin_role and (self.admin_role in user_roles)
+        self.log.info("Authentication Successful for user: %s, roles: %s, admin: %s" % (user['name'], user_roles, user['admin']))
+
+        return user
+
     async def pre_spawn_start(self, user, spawner):
-        auth_state = await user.get_auth_state()
-        spawner.access_token = auth_state['access_token']
-        spawner.user_roles = self._get_roles_for_token(auth_state['access_token'])
-        spawner.inspection_url = self.userdata_url
-        # If function raises Exception, let the this fail
-        await maybe_future(self.get_uid_hook(spawner, auth_state))
+        if self.pre_spawn_hook:
+            auth_state = await user.get_auth_state()
+            await maybe_future(self.pre_spawn_hook(self, spawner, auth_state))
 
 
     async def refresh_user(self, user, handler=None):
@@ -144,11 +191,13 @@ class KeyCloakAuthenticator(GenericOAuthenticator):
         try:
             # Retrieve user authentication info, decode, and check if refresh is needed
             auth_state = await user.get_auth_state()
+
             decoded_access_token = self._decode_token(auth_state['access_token'])
             decoded_refresh_token = self._decode_token(auth_state['refresh_token'])
 
             diff_access = decoded_access_token['exp'] - time.time()
-            diff_refresh = decoded_refresh_token['exp'] - time.time()
+            # If we request the offline_access scope, our refresh token won't have expiration
+            diff_refresh = (decoded_refresh_token['exp'] - time.time()) if 'exp' in decoded_refresh_token else 0
 
             if diff_access > self.auth_refresh_age:
                 # Access token is still valid and will stay until next refresh
@@ -160,32 +209,21 @@ class KeyCloakAuthenticator(GenericOAuthenticator):
 
             else:
                 # We need to refresh access token (which will also refresh the refresh token)
-                values = dict(
-                    grant_type = 'refresh_token',
-                    client_id = self.client_id,
-                    client_secret = self.client_secret,
-                    refresh_token = auth_state['refresh_token']
-                )
-                data = parse.urlencode(values).encode('ascii')
+                access_token, refresh_token = self._refresh_token(auth_state['refresh_token'])
+                auth_state['access_token'] = access_token
+                auth_state['refresh_token'] = refresh_token
+                auth_state['exchanged_tokens'] = self._exchange_tokens(access_token)
 
-                req = request.Request(self.token_url, data)
-                with request.urlopen(req) as response:
-                    data = json.loads(response.read())
-
-                    auth_state['access_token'] = data.get('access_token', None)
-                    auth_state['refresh_token'] = data.get('refresh_token', None)
-                    refresh_user_return = {
-                        'auth_state': auth_state
-                    }
-
-                    self.log.info('User %s oAuth tokens refreshed' % user.name)
-                    return refresh_user_return
+                self.log.info('User %s oAuth tokens refreshed' % user.name)
+                return {
+                    'auth_state': auth_state
+                }
 
         except HTTPError as e:
             self.log.error("Failure calling the renew endpoint: %s (code: %s)" % (e.read(), e.code))
 
         except:
-            self.log.error("Failed to refresh the oAuth token", exc_info=True)
+            self.log.error("Failed to refresh the oAuth tokens", exc_info=True)
 
         return False
 
